{-------------------------------------------}
{                 Real Mode                 }
{        Multitasking/multithreading        }
{                   Demo                    }
{                                           }
{       by Alexei A. Frounze (c) 2000       }
{                                           }
{ E-mail:   alexfru@chat.ru                 }
{ Homepage: http://alexfru.chat.ru          }
{ Mirror:   http://members.xoom.com/alexfru }
{-------------------------------------------}

USES CRT, DOS;

CONST
  STACK_SIZE = 4096;                            { TASK STACK SIZE }
  MAX_TASKS  = 8;                               { MAX TASKS }
  FASTER     = 8;                               { SPEED UP "PIT" BY "FASTER" }

VAR
  OLD_TIMER  : POINTER;
  SWITCHING  : BOOLEAN;                           { ENABLE/DISABLE SWITCHING }
  BLOCKS,                                         { STACK BUFFERS' ADDRESSES }
  STACK_LIST : ARRAY [0..MAX_TASKS-1] OF POINTER; { STACK POINTERS FOR TASKS }
  CUR_TASK   : WORD;                              { CURRENT TASK NO. }
  TASK_CNT   : WORD;                              { TASKS TOTAL }
  ACTIVE_TIME,                                    { ACTIVITY IN TICKS }
  AVAIL_TIME : ARRAY [0..MAX_TASKS-1] OF WORD;    { QUANT OF TIME IN TICS }

{$L MULTIT}
PROCEDURE SHEDULER; FAR; EXTERNAL;
PROCEDURE ADD_TASK (IP, CS, DS, AVAILTIME : WORD); FAR; EXTERNAL;

PROCEDURE INIT;
VAR I : WORD;
BEGIN
  SWITCHING := FALSE;
  TASK_CNT := 1;
  CUR_TASK := 0;
  FOR I := 0 TO MAX_TASKS-1 DO
    GETMEM (BLOCKS[I], STACK_SIZE);
  AVAIL_TIME[0] := 7;                   { 70% OF CPU TIME FOR MAIN PROGRAM }
  ACTIVE_TIME[0] := 0;
  GETINTVEC (8, OLD_TIMER);
  SETINTVEC (8, @SHEDULER);
  { REPROGRAMMING "PIT" FOR MORE SMOOTH THREAD EXECUTION }
  { SPEEDING UP FOR "FASTER" TIMES }
  PORT[$43] := $34;
  PORT[$40] := ($10000 DIV FASTER) AND $FF;
  PORT[$40] := (($10000 DIV FASTER) SHR 8) AND $FF;
END;

PROCEDURE DONE;
VAR I : WORD;
BEGIN
  SWITCHING := FALSE;
  FOR I := 0 TO MAX_TASKS-1 DO
    FREEMEM (BLOCKS[I], STACK_SIZE);
  SETINTVEC (8, OLD_TIMER);
  { REPROGRAMMING "PIT" BACK }
  { SETTING STANDARD 18.2 HZ FREQUENCY }
  PORT[$43] := $34;
  PORT[$40] := $00;
  PORT[$40] := $00;
END;

                               (* TASKS' CODE *)

{ We use this subroutine for screen output because CRT is a global unit }
{ and it wouldn't work properly, if every thread/task used it. }
{ In other words, DOS and some BP functions/procedures are not reenterable. }
procedure WriteStrXY (x, y : word; s : string);
var
  l, i : integer;
  addr : word;
begin
  i := 1;
  l := length(s);
  addr := (y*80+x) shl 1;
  while l > 0 do begin
    Mem[$B800:addr] := byte(s[i]);
    inc (addr, 2);
    inc (i);
    dec (l)
  end
end;

procedure task1;
var
  counter : longint;
  s       : string;
begin
  counter := 0;
  repeat
    str (counter, s);
    WriteStrXY (0, 0, 'Task 1 counter: '+s);
    inc (counter)
  until false
end;

procedure task2;
var
  counter : longint;
  s       : string;
begin
  counter := 0;
  repeat
    str (counter, s);
    WriteStrXY (0, 1, 'Task 2 counter: '+s);
    inc (counter)
  until false
end;

                              (* MAIN PROGRAM *)

BEGIN
  CLRSCR;
  INIT;                                 { 70% OF CPU TIME FOR MAIN PROGRAM }

  ADD_TASK (OFS(TASK1), CSEG, DSEG, 1); { 10% OF CPU TIME FOR TASK 1 }
  ADD_TASK (OFS(TASK2), CSEG, DSEG, 2); { 20% OF CPU TIME FOR TASK 2 }

  GOTOXY (1, 4); WRITELN ('HIT A KEY TO START SWITCHING...'); READKEY;

  SWITCHING := TRUE;

  WRITELN ('HIT A KEY TO SHUTDOWN...'#13#10); READKEY;

  DONE;

  WRITELN ('Ŀ');
  WRITELN ('                Real Mode                ');
  WRITELN ('       Multitasking/multithreading       ');
  WRITELN ('                  Demo                   ');
  WRITELN ('                                         ');
  WRITELN ('      by Alexei A. Frounze (c) 2000      ');
  WRITELN ('                                         ');
  WRITELN ('E-mail:   alexfru@chat.ru                ');
  WRITELN ('Homepage: http://alexfru.chat.ru         ');
  WRITELN ('Mirror:   http://members.xoom.com/alexfru');
  WRITELN ('');
END.
